import numpy as np
import pandas as pd
from tqdm import tqdm

from viprs.model.VIPRS import VIPRS
from ..vi.e_step import e_step_grid
from ..vi.e_step_cpp import cpp_e_step_grid
from viprs.utils.compute_utils import dict_mean
from viprs.utils.OptimizeResult import OptimizeResult


class VIPRSGrid(VIPRS):
    """
    A class to fit the `VIPRS` model to data using a grid of hyperparameters.
    Instead of having a single set of hyperparameters, we simultaneously fit
    multiple models with different hyperparameters and compare their performance
    at the end. This class is generic and does not support any model selection or
    averaging schemes.

    The class inherits all the basic attributes from the [VIPRS][viprs.model.VIPRS.VIPRS] class.

    !!! seealso "See Also"
        * [VIPRSGridSearch][viprs.model.gridsearch.VIPRSGridSearch.VIPRSGridSearch]
        * [VIPRSBMA][viprs.model.gridsearch.VIPRSBMA.VIPRSBMA]

    :ivar grid_table: A pandas table containing the hyperparameters for each model.
    :ivar n_models: The number of models to fit.
    :ivar shapes: A dictionary containing the shapes of the data matrices.
    :ivar active_models: A boolean array indicating which models are still active (i.e. not converged).

    """

    def __init__(self,
                 gdl,
                 grid,
                 **kwargs):
        """
        Initialize the `VIPRS` model with a grid of hyperparameters.

        :param gdl: An instance of `GWADataLoader`
        :param grid: An instance of `HyperparameterGrid`
        :param kwargs: Additional keyword arguments for the `VIPRS` model
        """

        self.grid_table = grid.to_table()
        self.n_models = len(self.grid_table)
        assert self.n_models > 1

        grid_params = {c: self.grid_table[c].values for c in self.grid_table.columns}

        if 'fix_params' not in kwargs:
            kwargs['fix_params'] = grid_params
        else:
            kwargs['fix_params'].update(grid_params)

        # Make sure that the matrices are in Fortran order:
        kwargs['order'] = 'F'

        super().__init__(gdl, **kwargs)

        self.shapes = {c: (shp, self.n_models)
                       for c, shp in self.shapes.items()}
        self.active_models = None
        self.Nj = {c: Nj[:, None].astype(self.float_precision, order=self.order) for c, Nj in self.Nj.items()}
        self.optim_results = [OptimizeResult() for _ in range(self.n_models)]

    @property
    def models_to_keep(self):
        """
        :return: A boolean array indicating which models have converged successfully.
        """
        return np.logical_or(self.active_models, [optr.success for optr in self.optim_results])

    def initialize_theta(self, theta_0=None):
        """
        Initialize the global hyperparameters of the model.
        :param theta_0: A dictionary of initial values for the hyperparameters theta
        """

        self.active_models = np.array([True for _ in range(self.n_models)])
        for optr in self.optim_results:
            optr.reset()

        super().initialize_theta(theta_0=theta_0)

        try:
            if self.pi.shape != (self.n_models, ):
                self.pi *= np.ones(shape=(self.n_models, ), dtype=self.float_precision)
        except AttributeError:
            self.pi *= np.ones(shape=(self.n_models,), dtype=self.float_precision)

        try:
            if self.tau_beta.shape != (self.n_models, ):
                self.tau_beta *= np.ones(shape=(self.n_models, ), dtype=self.float_precision)
        except AttributeError:
            self.tau_beta *= np.ones(shape=(self.n_models,), dtype=self.float_precision)

        try:
            if self.sigma_epsilon.shape != (self.n_models, ):
                self.sigma_epsilon *= np.ones(shape=(self.n_models, ), dtype=self.float_precision)
        except AttributeError:
            self.sigma_epsilon *= np.ones(shape=(self.n_models,), dtype=self.float_precision)

    def e_step(self):
        """
        Run the E-Step of the Variational EM algorithm.
        Here, we update the variational parameters for each variant using coordinate
        ascent optimization techniques. The coordinate ascent procedure is run on all the models
        in the grid simultaneously. The update equations are outlined in
        the Supplementary Material of the following paper:

        > Zabad S, Gravel S, Li Y. Fast and accurate Bayesian polygenic risk modeling with variational inference.
        Am J Hum Genet. 2023 May 4;110(5):741-761. doi: 10.1016/j.ajhg.2023.03.009.
        Epub 2023 Apr 7. PMID: 37030289; PMCID: PMC10183379.
        """

        active_model_idx = np.where(self.active_models)[0].astype(np.int32)

        for c, shapes in self.shapes.items():

            # Get the priors:
            tau_beta = self.get_tau_beta(c)
            pi = self.get_pi(c)

            # Updates for tau variational parameters:
            # NOTE: Here, we compute the variational sigma in-place to avoid the need
            # to change the order of the resulting matrix or its float precision:
            np.add(self.Nj[c] / self.sigma_epsilon, tau_beta, out=self.var_tau[c])

            # Compute some quantities that are needed for the per-SNP updates:
            mu_mult = self.Nj[c] / (self.var_tau[c] * self.sigma_epsilon)
            u_logs = np.log(pi) - np.log(1. - pi) + .5 * (np.log(tau_beta) - np.log(self.var_tau[c]))

            if self.use_cpp:
                cpp_e_step_grid(self.ld_left_bound[c],
                                self.ld_indptr[c],
                                self.ld_data[c],
                                self.std_beta[c],
                                self.var_gamma[c],
                                self.var_mu[c],
                                self.eta[c],
                                self.q[c],
                                self.eta_diff[c],
                                u_logs,
                                0.5 * self.var_tau[c],
                                mu_mult,
                                self.dequantize_scale,
                                active_model_idx,
                                self.threads,
                                self.use_blas,
                                self.low_memory)
            else:

                e_step_grid(self.ld_left_bound[c],
                            self.ld_indptr[c],
                            self.ld_data[c],
                            self.std_beta[c],
                            self.var_gamma[c],
                            self.var_mu[c],
                            self.eta[c],
                            self.q[c],
                            self.eta_diff[c],
                            u_logs,
                            0.5 * self.var_tau[c],
                            mu_mult,
                            active_model_idx,
                            self.threads,
                            self.use_blas,
                            self.low_memory)

        self.zeta = self.compute_zeta()

    def to_theta_table(self):
        """
        :return: A `pandas` DataFrame containing information about the hyperparameters of the model.
        """

        if self.n_models == 1:
            return super(VIPRSGrid, self).to_theta_table()

        sig_e = self.sigma_epsilon
        h2 = self.get_heritability()
        pi = self.get_proportion_causal()

        if isinstance(self.tau_beta, dict):
            taus = dict_mean(self.tau_beta, axis=0)
        else:
            taus = self.tau_beta

        theta_table = []

        for m in range(self.n_models):

            theta_table += [
                {'Model': m, 'Parameter': 'Residual_variance', 'Value': sig_e[m]},
                {'Model': m, 'Parameter': 'Heritability', 'Value': h2[m]},
                {'Model': m, 'Parameter': 'Proportion_causal', 'Value': pi[m]},
                {'Model': m, 'Parameter': 'sigma_beta', 'Value': taus[m]}
            ]

        return pd.DataFrame(theta_table)

    def to_validation_table(self):
        """
        :return: The validation table summarizing the performance of each model.
        :raises ValueError: if the validation result is not set.
        """
        if self.validation_result is None:
            raise ValueError("Validation result is not set!")
        elif len(self.validation_result) < 1:
            raise ValueError("Validation result is not set!")

        return pd.DataFrame(self.validation_result)

    def write_validation_result(self, v_filename, sep="\t"):
        """
        After performing hyperparameter search, write a table
        that records that value of the objective for each combination
        of hyperparameters.
        :param v_filename: The filename for the validation table.
        :param sep: The separator for the validation table
        """

        v_df = self.to_validation_table()
        v_df.to_csv(v_filename, index=False, sep=sep)

    def fit(self,
            max_iter=1000,
            theta_0=None,
            param_0=None,
            continued=False,
            f_abs_tol=1e-6,
            x_abs_tol=1e-6,
            drop_r_tol=1e-6,
            patience=5):
        """
        A convenience method to fit all the models in the grid using the Variational EM algorithm.

        :param max_iter: Maximum number of iterations. 
        :param theta_0: A dictionary of values to initialize the hyperparameters
        :param param_0: A dictionary of values to initialize the variational parameters
        :param continued: If true, continue the model fitting for more iterations.
        :param f_abs_tol: The absolute tolerance threshold for the objective (ELBO).
        :param x_abs_tol: The absolute tolerance threshold for the variational parameters.
        :param drop_r_tol: The relative tolerance for the drop in the ELBO to be considered as a red flag. It usually
        happens around convergence that the objective fluctuates due to numerical errors. This is a way to
        differentiate such random fluctuations from actual drops in the objective.
        :param patience: The maximum number of times the objective is allowed to drop before termination.
        """

        if not continued:
            self.initialize(theta_0, param_0)
            start_idx = 1
        else:
            start_idx = len(self.history['ELBO']) + 1

        patience = patience*np.ones(self.n_models)

        if int(self.verbose) > 1:
            print("> Performing model fit...")
            if self.threads > 1:
                print(f"> Using up to {self.threads} threads.")

        # Progress bar:
        pbar = tqdm(range(start_idx, start_idx + max_iter),
                    disable=not self.verbose)

        for i in pbar:

            if all([optr.stop_iteration for optr in self.optim_results]):

                # If converged, update the progress bar before exiting:
                pbar.set_postfix({'Best ELBO': f"{self.history['ELBO'][-1][self.models_to_keep].max():.4f}",
                                  'Models converged': f"{self.n_models - np.sum(self.active_models)}/{self.n_models}"})
                pbar.n = i - 1
                pbar.total = i - 1
                pbar.refresh()
                pbar.close()

                pbar.close()
                break

            self.update_theta_history()

            self.e_step()
            self.m_step()

            self.history['ELBO'].append(self.elbo(sum_axis=0))
            h2 = self.get_heritability()

            if i > 1:

                # Get the current and previous ELBO values
                curr_elbo = self.history['ELBO'][-1]
                prev_elbo = self.history['ELBO'][-2]

                for m, active in enumerate(self.active_models):
                    if active:

                        if np.isclose(prev_elbo[m], curr_elbo[m], atol=f_abs_tol, rtol=0.):
                            self.active_models[m] = False
                            self.optim_results[m].update(curr_elbo[m],
                                                         stop_iteration=True,
                                                         success=True,
                                                         message='Objective (ELBO) converged successfully.')

                        # Check to see if the objective drops due to numerical instabilities:
                        elif curr_elbo[m] < prev_elbo[m] and not np.isclose(curr_elbo[m],
                                                                            prev_elbo[m],
                                                                            atol=0.,
                                                                            rtol=drop_r_tol):
                            patience[m] -= 1

                            if patience[m] == 0:
                                self.active_models[m] = False
                                self.optim_results[m].update(curr_elbo[m],
                                                             stop_iteration=True,
                                                             success=False,
                                                             message='Optimization is halted '
                                                                     'due to numerical instabilities.')
                            else:
                                self.optim_results[m].update(curr_elbo)

                        # Check if the model parameters behave in unexpected/pathological ways:
                        elif np.isnan(curr_elbo[m]):
                            self.active_models[m] = False
                            self.optim_results[m].update(curr_elbo[m],
                                                         stop_iteration=True,
                                                         success=False,
                                                         message='The objective (ELBO) is NaN.')
                        elif self.sigma_epsilon[m] <= 0.:
                            self.active_models[m] = False
                            self.optim_results[m].update(curr_elbo[m],
                                                         stop_iteration=True,
                                                         success=False,
                                                         message='Optimization is halted (sigma_epsilon <= 0).')
                        elif h2[m] >= 1.:
                            self.active_models[m] = False
                            self.optim_results[m].update(curr_elbo[m],
                                                         stop_iteration=True,
                                                         success=False,
                                                         message='Optimization is halted (h2 >= 1).')
                        else:
                            self.optim_results[m].update(curr_elbo)

                # -----------------------------------------------------------------------

            if self.models_to_keep.sum() > 0:
                pbar.set_postfix({'Best ELBO': f"{self.history['ELBO'][-1][self.models_to_keep].max():.4f}",
                                  'Models converged': f"{self.n_models - np.sum(self.active_models)}/{self.n_models}"})

                # TODO: Figure out how to update only models that improved.
                self.update_posterior_moments()

        for m, optr in enumerate(self.optim_results):
            if not optr.stop_iteration:
                self.active_models[m] = False
                optr.update(self.history['ELBO'][-1][m],
                            stop_iteration=True,
                            success=False,
                            message="Maximum iterations reached without convergence.\n"
                                    "You may need to run the model for more iterations.")

        if int(self.verbose) > 1:

            if self.models_to_keep.sum() > 0:
                print(f"> Optimization is complete for all {self.n_models} models.")
                print(f"> {np.sum(self.models_to_keep)} model(s) converged successfully.")
            else:
                print("> All models failed to converge. Please check the optimization results.")

        self.validation_result = self.grid_table.copy()
        self.validation_result['ELBO'] = [optr.fun for optr in self.optim_results]
        self.validation_result['Converged'] = self.models_to_keep
        self.validation_result['Optimization_message'] = [optr.message for optr in self.optim_results]

        return self



import pandas as pd
import seaborn as sns


def plot_history(prs_model, quantity=None):
    """
    This function plots the optimization history for various model parameters and/or objectives. For
    every iteration step, we generally save quantities such as the ELBO, the heritability, etc. For the purposes
    of debugging and checking model convergence, it is useful to visually observe the trajectory
    of these quantities as a function of training iteration.

    :param prs_model: A `VIPRS` (or its derived classes) object.
    :param quantity: The quantities to plot (e.g. `ELBO`, `heritability`, etc.).
    """

    if quantity is None:
        quantity = prs_model.history.keys()
    elif isinstance(quantity, str):
        quantity = [quantity]

    q_dfs = []

    for attr in quantity:

        df = pd.DataFrame({'Value': prs_model.history[attr]})
        df.reset_index(inplace=True)
        df.columns = ['Step', 'Value']
        df['Quantity'] = attr

        q_dfs.append(df)

    q_dfs = pd.concat(q_dfs)

    g = sns.relplot(
        data=q_dfs, x="Step", y="Value",
        row="Quantity",
        facet_kws={'sharey': False, 'sharex': True},
        kind="scatter",
        marker="."
    )

    return g



import numpy as np
import pandas as pd
from scipy import stats
import statsmodels.api as sm


def r2_stats(r2_val, n):
    """
    Compute the confidence interval and p-value for a given R-squared (proportion of variance
     explained) value.

    This function and the formulas therein are based on the following paper
    by Momin et al. 2023: https://doi.org/10.1016/j.ajhg.2023.01.004 as well as
    the implementation in the R package `PRSmix`:
    https://github.com/buutrg/PRSmix/blob/main/R/get_PRS_acc.R#L63

    :param r2_val: The R^2 value to compute the confidence interval/p-value for.
    :param n: The sample size used to compute the R^2 value

    :return: A dictionary with the R^2 value, the lower and upper values of the confidence interval,
    the p-value, and the standard error of the R^2 metric.

    """

    assert 0. < r2_val < 1., "R^2 value must be between 0 and 1."

    # Compute the variance of the R^2 value:
    r2_var = (4. * r2_val * (1. - r2_val) ** 2 * (n - 2) ** 2) / ((n ** 2 - 1) * (n + 3))

    # Compute the standard errors for the R^2 value
    # as well as the lower and upper values for
    # the confidence interval:
    r2_se = np.sqrt(r2_var)
    lower_r2 = r2_val - 1.97 * r2_se
    upper_r2 = r2_val + 1.97 * r2_se

    # Compute the p-value assuming a Chi-squared distribution with 1 degree of freedom:
    pval = stats.chi2.sf((r2_val / r2_se) ** 2, df=1)

    return {
        'R2': r2,
        'Lower_R2': lower_r2,
        'Upper_R2': upper_r2,
        'P_Value': pval,
        'SE': r2_se,
    }


def r2(true_val, pred_val):
    """
    Compute the R^2 (proportion of variance explained) between
    the predictions or PRS `pred_val` and the phenotype `true_val`
    
    :param true_val: The response value or phenotype (a numpy vector)
    :param pred_val: The predicted value or PRS (a numpy vector)
    """
    _, _, r_val, _, _ = stats.linregress(pred_val, true_val)
    return r_val ** 2


def mse(true_val, pred_val):
    """
    Compute the mean squared error (MSE) between
    the predictions or PRS `pred_val` and the phenotype `true_val`
    
    :param true_val: The response value or phenotype (a numpy vector)
    :param pred_val: The predicted value or PRS (a numpy vector)
    """

    return np.mean((pred_val - true_val)**2)


def pearson_r(true_val, pred_val):
    """
    Compute the pearson correlation coefficient between
    the predictions or PRS `pred_val` and the phenotype `true_val`
    
    :param true_val: The response value or phenotype (a numpy vector)
    :param pred_val: The predicted value or PRS (a numpy vector)
    """
    return np.corrcoef(true_val, pred_val)[0, 1]


def incremental_r2(true_val, pred_val, covariates=None, return_all_r2=False):
    """
    Compute the incremental prediction R^2 (proportion of phenotypic variance explained by the PRS).
    This metric is computed by taking the R^2 of a model with covariates+PRS and subtracting from it
    the R^2 of a model with covariates alone covariates.

    :param true_val: The response value or phenotype (a numpy vector)
    :param pred_val: The predicted value or PRS (a numpy vector)
    :param covariates: A pandas table of covariates where the rows are ordered
    the same way as the predictions and response.
    :param return_all_r2: If True, return the R^2 values for the null and full models as well.
    """

    if covariates is None:
        covariates = pd.DataFrame(np.ones((true_val.shape[0], 1)), columns=['const'])
    else:
        covariates = sm.add_constant(covariates)

    null_result = sm.OLS(true_val, covariates).fit(disp=0)
    full_result = sm.OLS(true_val, covariates.assign(pred_val=pred_val)).fit(disp=0)

    if return_all_r2:
        return {
            'Null_R2': null_result.rsquared,
            'Full_R2': full_result.rsquared,
            'Incremental_R2': full_result.rsquared - null_result.rsquared
        }
    else:
        return full_result.rsquared - null_result.rsquared


def partial_correlation(true_val, pred_val, covariates):
    """
    Compute the partial correlation between the phenotype `true_val` and the PRS `pred_val`
    by conditioning on a set of covariates. This metric is computed by first residualizing the
    phenotype and the PRS on a set of covariates and then computing the correlation coefficient
    between the residuals.

    :param true_val: The response value or phenotype (a numpy vector)
    :param pred_val: The predicted value or PRS (a numpy vector)
    :param covariates: A pandas table of covariates where the rows are ordered
    the same way as the predictions and response.
    """

    true_response = sm.OLS(true_val, sm.add_constant(covariates)).fit(disp=0)
    pred_response = sm.OLS(pred_val, sm.add_constant(covariates)).fit(disp=0)

    return np.corrcoef(true_response.resid, pred_response.resid)[0, 1]



import pandas as pd
import numpy as np
import copy
from tqdm import tqdm
import multiprocessing
from pprint import pprint

from viprs.eval.continuous_metrics import r2
from viprs.eval.binary_metrics import roc_auc
from viprs.model.BayesPRSModel import BayesPRSModel
from viprs.model.VIPRS import VIPRS


def fit_model_fixed_params(params):
    """
    Perform model fitting using a set of fixed parameters.
    This is a helper function to allow us to use the `multiprocessing` module
    to fit PRS models in parallel.
    :param params: A tuple of (BayesPRSModel, fixed parameters dictionary, and kwargs for the .fit() method).
    """

    # vi_model, fixed_params, **fit_kwargs
    vi_model, fixed_params, fit_kwargs = params
    vi_model.fix_params = fixed_params

    try:
        vi_model.fit(**fit_kwargs)
    except Exception as e:
        return None

    return vi_model


class HyperparameterSearch(object):
    """
    A generic class for performing hyperparameter search on the
    `VIPRS` model. This interface is old and will likely be deprecated
    in future releases. It is recommended to use the `VIPRSGrid` class
    and its derivatives for performing grid search instead.
    """

    def __init__(self,
                 gdl,
                 model=None,
                 criterion='ELBO',
                 validation_gdl=None,
                 verbose=False,
                 n_jobs=1):
        """
        A generic hyperparameter search class that implements common functionalities
        that may be required by hyperparameter search strategies.
        :param gdl: A GWADataLoader object
        :param model: A `PRSModel`-derived object (e.g. VIPRS).
        :param criterion: The objective function for the hyperparameter search.
        Options are: `ELBO`, `pseudo_validation` or `validation`.
        :param validation_gdl: If the objective is validation, provide the GWADataLoader object for the validation
        dataset.
        :param verbose: Detailed messages and print statements.
        :param n_jobs: The number of processes to use for the hyperparameters search.
        """

        # Sanity checking:
        assert criterion in ('ELBO', 'validation', 'pseudo_validation')

        self.gdl = gdl
        self.n_jobs = n_jobs

        if model is None:
            self.model = VIPRS(gdl)
        else:
            import inspect
            if inspect.isclass(model):
                self.model = model(gdl)
            else:
                self.model = model

        self.validation_result = None

        self.criterion = criterion
        self._validation_gdl = validation_gdl

        self.verbose = verbose
        self.model.verbose = verbose

        if self._validation_gdl is not None:
            self._validation_gdl.verbose = verbose

        if self.criterion == 'ELBO':
            assert hasattr(self.model, 'elbo')
        elif self.criterion == 'pseudo_validation':
            assert self._validation_gdl is not None
            assert self._validation_gdl.sumstats_table is not None
        if self.criterion == 'validation':
            assert self._validation_gdl is not None
            assert self._validation_gdl.genotype is not None
            assert self._validation_gdl.sample_table.phenotype is not None

    def to_validation_table(self):
        """
        Summarize the validation results in a pandas table.
        """
        if self.validation_result is None:
            raise Exception("Validation result is not set!")
        elif len(self.validation_result) < 1:
            raise Exception("Validation result is not set!")

        return pd.DataFrame(self.validation_result)

    def write_validation_result(self, v_filename, sep="\t"):
        """
        After performing hyperparameter search, write a table
        that records that value of the objective for each combination
        of hyperparameters.
        :param v_filename: The filename for the validation table.
        :param sep: The separator for the validation table
        """

        v_df = self.to_validation_table()
        v_df.to_csv(v_filename, index=False, sep=sep)

    def multi_objective(self, models):
        """
        This method evaluates multiple PRS models simultaneously. This can be faster for
        some evaluation criteria, such as the validation R^2, because we only need to
        multiply the inferred effect sizes with the genotype matrix only once.

        :param models: A list of PRS models that we wish to evaluate.
        """

        if len(models) == 1:
            return self.objective(models[0])

        if self.criterion == 'ELBO':
            return [m.elbo() for m in models]

        elif self.criterion == 'pseudo_validation':
            return [m.pseudo_validate(validation_gdl=self._validation_gdl) for m in models]
        else:

            prs_m = BayesPRSModel(self._validation_gdl)

            eff_table = models[0].to_table(per_chromosome=False)
            eff_table = eff_table[['CHR', 'SNP', 'A1', 'A2', 'BETA']]
            eff_table.rename(columns={'BETA': 'BETA_0'}, inplace=True)

            eff_table[[f'BETA_{i}' for i in range(1, len(models))]] = np.array(
                [models[i].to_table(per_chromosome=False)['BETA'].values for i in range(1, len(models))]
            ).T

            prs_m.set_model_parameters(eff_table)

            prs = prs_m.predict(test_gdl=self._validation_gdl)

            if self._validation_gdl.phenotype_likelihood == 'binomial':
                eval_func = roc_auc
            else:
                eval_func = r2

            metrics = [eval_func(prs[:, i].flatten(), self._validation_gdl.sample_table.phenotype)
                       for i in range(len(models))]

            return metrics

    def objective(self, model):
        """
        A method that takes the result of fitting the model
        and returns the desired objective (either `ELBO`, `pseudo_validation`, or `validation`).
        :param model: The PRS model to evaluate
        """

        if self.criterion == 'ELBO':
            return model.elbo()
        elif self.criterion == 'pseudo_validation':
            return model.pseudo_validate(validation_gdl=self._validation_gdl)
        else:

            # Predict:
            prs = model.predict(test_gdl=self._validation_gdl)

            if self._validation_gdl.phenotype_likelihood == 'binomial':
                eval_func = roc_auc
            else:
                eval_func = r2

            return eval_func(prs, self._validation_gdl.sample_table.phenotype)

    def fit(self):
        raise NotImplementedError


class BayesOpt(HyperparameterSearch):
    """
    Hyperparameter search using Bayesian optimization
    """

    def __init__(self,
                 gdl,
                 opt_params,
                 param_bounds=None,
                 model=None,
                 criterion='ELBO',
                 validation_gdl=None,
                 verbose=False,
                 n_jobs=1):
        """
        Perform hyperparameter search using Bayesian optimization
        :param gdl: A GWADataLoader object
        :param opt_params: A list of the hyperparameters to optimize over (e.g. 'pi', 'sigma_epsilon', 'sigma_beta').
        :param param_bounds: The bounds for each hyperparameter included in the optimization. A list of tuples,
        where each tuples records the (min, max) values for each hyperparameter.
        :param model: A `PRSModel`-derived object (e.g. VIPRS).
        :param criterion: The objective function for the hyperparameter search (ELBO or validation).
        :param validation_gdl: If the objective is validation, provide the GWADataLoader object for the validation
        dataset.
        :param verbose: Detailed messages and print statements.
        :param n_jobs: The number of processes to use for the hyperparameters search (not applicable here).
        """
        
        super().__init__(gdl,
                         model=model,
                         criterion=criterion,
                         validation_gdl=validation_gdl,
                         verbose=verbose,
                         n_jobs=n_jobs)

        self._opt_params = opt_params
        self._param_bounds = param_bounds or {
            'sigma_epsilon': (1e-6, 1. - 1e-6),
            'tau_beta': (1e-3, None),
            'pi': (1e-6, 1. - 1e-6)
        }

        # Convert the `pi` limits to log-scale:
        if 'pi' in self._param_bounds:
            self._param_bounds['pi'] = tuple(np.log10(list(self._param_bounds['pi'])))

        assert all([opp in self._param_bounds for opp in self._opt_params])

    def fit(self,
            max_iter=50,
            f_abs_tol=1e-4,
            n_calls=30,
            n_random_starts=5,
            acq_func="gp_hedge"):
        """
        Perform model fitting and hyperparameter search using Bayesian optimization.

        :param n_calls: The number of model runs with different hyperparameter settings.
        :param n_random_starts: The number of random starts to initialize the optimizer.
        :param acq_func: The acquisition function (default: `gp_hedge`)
        :param max_iter: The maximum number of iterations within the search (default: 50).
        :param f_abs_tol: The absolute tolerance for the objective (ELBO) within the search
        """

        from skopt import gp_minimize

        def opt_func(p):

            fix_params = dict(zip(self._opt_params, p))
            if 'pi' in fix_params:
                fix_params['pi'] = 10**fix_params['pi']

            fitted_model = fit_model_fixed_params((self.model, fix_params,
                                                   {'max_iter': max_iter,
                                                    'f_abs_tol': f_abs_tol}))

            if fitted_model is None:
                return np.inf
            else:
                return -self.objective(fitted_model)

        res = gp_minimize(opt_func,  # the function to minimize
                          [self._param_bounds[op] for op in self._opt_params],  # the bounds on each dimension of x
                          acq_func=acq_func,  # the acquisition function
                          n_calls=n_calls,  # the number of evaluations of f
                          n_random_starts=n_random_starts)  # the random seed

        # Store validation result
        self.validation_result = []
        for obj, x in zip(res.func_vals, res.x_iters):
            v_res = dict(zip(self._opt_params, x))
            if 'pi' in v_res:
                v_res['pi'] = 10**v_res['pi']

            if self.criterion == 'ELBO':
                v_res['ELBO'] = -obj
            elif self.criterion == 'pseudo_validation':
                v_res['Pseudo_Validation_Corr'] = -obj
            else:
                v_res['Validation_R2'] = -obj

            self.validation_result.append(v_res)

        # Extract the best performing hyperparameters:
        final_best_params = dict(zip(self._opt_params, res.x))
        if 'pi' in final_best_params:
            final_best_params['pi'] = 10 ** final_best_params['pi']

        print("> Bayesian Optimization identified the best hyperparameters as:")
        pprint(final_best_params)

        print("> Refitting the model with the best hyperparameters...")

        self.model.fix_params = final_best_params
        return self.model.fit()


class GridSearch(HyperparameterSearch):
    """
    Hyperparameter search using Grid Search
    """

    def __init__(self,
                 gdl,
                 grid,
                 model=None,
                 criterion='ELBO',
                 validation_gdl=None,
                 verbose=False,
                 n_jobs=1):

        """
        Perform hyperparameter search using grid search
        :param gdl: A GWADataLoader object
        :param grid: A HyperParameterGrid object
        :param model: A `PRSModel`-derived object (e.g. VIPRS).
        :param criterion: The objective function for the grid search (ELBO or validation).
        :param validation_gdl: If the objective is validation, provide the GWADataLoader object for the validation
        dataset.
        :param verbose: Detailed messages and print statements.
        :param n_jobs: The number of processes to use for the grid search
        """

        super().__init__(gdl, model=model, criterion=criterion,
                         validation_gdl=validation_gdl,
                         verbose=verbose,
                         n_jobs=n_jobs)

        self.grid = grid
        self.model.threads = 1

    def fit(self, max_iter=50, f_abs_tol=1e-3, x_abs_tol=1e-8):

        print("> Performing Grid Search over the following grid:")
        print(self.grid.to_table())

        opts = [(self.model, g, {'max_iter': max_iter,
                                 'f_abs_tol': f_abs_tol,
                                 'x_abs_tol': x_abs_tol})
                for g in self.grid.combine_grids()]

        assert len(opts) > 1

        self.validation_result = []
        fit_results = []
        params = []

        ctx = multiprocessing.get_context("spawn")

        with ctx.Pool(self.n_jobs, maxtasksperchild=1) as pool:

            for idx, fitted_model in tqdm(enumerate(pool.imap(fit_model_fixed_params, opts)), total=len(opts)):

                if fitted_model is None:
                    continue

                fit_results.append(fitted_model)
                params.append(copy.copy(opts[idx][1]))
                self.validation_result.append(copy.copy(opts[idx][1]))
                self.validation_result[-1]['ELBO'] = fitted_model.elbo()

        if len(fit_results) > 1:
            res_objectives = self.multi_objective(fit_results)
        else:
            raise Exception("Error: Convergence was achieved for less than 2 models.")

        if self.criterion == 'validation':
            for i in range(len(self.validation_result)):
                self.validation_result[i]['Validation_R2'] = res_objectives[i]
        elif self.criterion == 'pseudo_validation':
            for i in range(len(self.validation_result)):
                self.validation_result[i]['Pseudo_Validation_Corr'] = res_objectives[i]

        best_idx = np.argmax(res_objectives)
        best_params = params[best_idx]

        print("> Grid search identified the best hyperparameters as:")
        pprint(best_params)

        print("> Refitting the model with the best hyperparameters...")

        self.model.fix_params = best_params
        return self.model.fit()


class BMA(BayesPRSModel):
    """
    Bayesian Model Averaging fitting procedure
    """

    def __init__(self,
                 gdl,
                 grid,
                 model=None,
                 normalization='softmax',
                 verbose=False,
                 n_jobs=1):
        """
        Integrate out hyperparameters using Bayesian Model Averaging
        :param gdl: A GWADataLoader object
        :param grid: A HyperParameterGrid object
        :param model: A `PRSModel`-derived object (e.g. VIPRS).
        :param normalization: The normalization scheme for the final ELBOs. Options are (`softmax`, `sum`).
        :param verbose: Detailed messages and print statements.
        :param n_jobs: The number of processes to use for the BMA
        """

        super().__init__(gdl)

        assert normalization in ('softmax', 'sum')

        self.grid = grid
        self.n_jobs = n_jobs
        self.verbose = verbose

        if model is None:
            self.model = VIPRS(gdl)
        else:
            self.model = model

        self.model.verbose = verbose
        self.model.threads = 1

        self.normalization = normalization

        self.var_gamma = None
        self.var_mu = None
        self.var_sigma = None

    def initialize(self):

        self.var_gamma = {c: np.zeros(c_size) for c, c_size in self.shapes.items()}
        self.var_mu = {c: np.zeros(c_size) for c, c_size in self.shapes.items()}
        self.var_sigma = {c: np.zeros(c_size) for c, c_size in self.shapes.items()}

    def fit(self, max_iter=100, f_abs_tol=1e-3, x_abs_tol=1e-8, **grid_kwargs):

        self.initialize()

        print("> Performing Bayesian Model Averaging with the following grid:")
        print(self.grid.to_table())

        opts = [(self.model, g, {'max_iter': max_iter,
                                 'f_abs_tol': f_abs_tol,
                                 'x_abs_tol': x_abs_tol})
                for g in self.grid.combine_grids()]

        elbos = []
        var_gammas = []
        var_mus = []
        var_sigmas = []

        ctx = multiprocessing.get_context("spawn")

        with ctx.Pool(self.n_jobs, maxtasksperchild=1) as pool:
            for fitted_model in tqdm(pool.imap_unordered(fit_model_fixed_params, opts), total=len(opts)):

                if fitted_model is None:
                    continue

                elbos.append(fitted_model.elbo())
                var_gammas.append(fitted_model.var_gamma)
                var_mus.append(fitted_model.var_mu)
                var_sigmas.append(fitted_model.var_sigma)

        elbos = np.array(elbos)

        if self.normalization == 'softmax':
            from scipy.special import softmax
            elbos = softmax(elbos)
        elif self.normalization == 'sum':
            # Correction for negative ELBOs:
            elbos = elbos - elbos.min() + 1.
            elbos /= elbos.sum()

        for idx in range(len(elbos)):
            for c in self.shapes:
                self.var_gamma[c] += var_gammas[idx][c]*elbos[idx]
                self.var_mu[c] += var_mus[idx][c]*elbos[idx]
                self.var_sigma[c] += var_sigmas[idx][c]*elbos[idx]

        self.pip = {}
        self.post_mean_beta = {}
        self.post_var_beta = {}

        for c, v_gamma in self.var_gamma.items():

            if len(v_gamma.shape) > 1:
                self.pip[c] = v_gamma.sum(axis=1)
                self.post_mean_beta[c] = (v_gamma*self.var_mu[c]).sum(axis=1)
                self.post_var_beta[c] = ((v_gamma * (self.var_mu[c] ** 2 + self.var_sigma[c])).sum(axis=1) -
                                         self.post_mean_beta[c]**2)
            else:
                self.pip[c] = v_gamma
                self.post_mean_beta[c] = v_gamma * self.var_mu[c]
                self.post_var_beta[c] = (v_gamma * (self.var_mu[c] ** 2 + self.var_sigma[c]) -
                                         self.post_mean_beta[c]**2)

        return self



from .BayesPRSModel import BayesPRSModel


class LDPredInf(BayesPRSModel):
    """
    A wrapper class implementing the LDPred-inf model.
    The LDPred-inf model is a Bayesian model that uses summary statistics
    from GWAS to estimate the posterior mean effect sizes of the SNPs. It is equivalent
    to performing ridge regression, with the penalty proportional to the inverse of
    the per-SNP heritability.

    Refer to the following references for details about the LDPred-inf model:
    * Vilhjálmsson et al. AJHG. 2015
    * Privé et al. Bioinformatics. 2020

    :ivar gdl: An instance of `GWADataLoader`
    :ivar h2: The heritability for the trait (can also be chromosome-specific)

    """

    def __init__(self,
                 gdl,
                 h2=None):
        """
        Initialize the LDPred-inf model.
        :param gdl: An instance of GWADataLoader
        :param h2: The heritability for the trait (can also be chromosome-specific)
        """
        super().__init__(gdl)

        if h2 is None:
            from magenpy.stats.h2.ldsc import simple_ldsc
            self.h2 = simple_ldsc(self.gdl)
        else:
            self.h2 = h2

    def get_heritability(self):
        """
        :return: The heritability estimate for the trait of interest.
        """
        return self.h2

    def fit(self, solver='minres', **solver_kwargs):
        """
        Fit the summary statistics-based ridge regression,
        following the specifications of the LDPred-inf model.

        !!! warning
            Not tested yet.

        Here, we use `lsqr` or `minres` solvers to solve the system of equations:

        (D + lam*I)BETA = BETA_HAT

        where D is the LD matrix, BETA is ridge regression
        estimate that we wish to obtain and BETA_HAT is the
        marginal effect sizes estimated from GWAS.

        In this case, lam = M / N*h2, where M is the number of SNPs,
        N is the number of samples and h2 is the heritability
        of the trait.

        https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsqr.html
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.minres.html

        :param solver: The solver for the system of linear equations. Options: `minres` or `lsqr`
        :param solver_kwargs: keyword arguments for the solver.
        """

        assert solver in ('lsqr', 'minres')

        import numpy as np
        from scipy.sparse.linalg import lsqr, minres
        from scipy.sparse import identity, block_diag

        if solver == 'lsqr':
            solve = lsqr
        else:
            solve = minres

        # Lambda, the regularization parameter for the
        # ridge regression estimator. For LDPred-inf model,
        # we set this to 'M / N*h2', where M is the number of SNPs,
        # N is the number of samples and h2 is the heritability
        # of the trait.
        lam = self.n_snps / (self.n * self.h2)

        chroms = self.gdl.chromosomes

        # Extract the LD matrices for all the chromosomes represented and
        # concatenate them into one block diagonal matrix:
        ld_mats = []
        for c in chroms:
            self.gdl.ld[c].load(dtype=np.float32)
            ld_mats.append(self.gdl.ld[c].csr_matrix)

        ld = block_diag(ld_mats, format='csr')

        # Extract the marginal GWAS effect sizes:
        marginal_beta = np.concatenate([self.gdl.sumstats_table[c].marginal_beta
                                        for c in chroms])

        # Estimate the BETAs under the ridge penalty:
        res = solve(ld + lam * identity(ld.shape[0]), marginal_beta, **solver_kwargs)

        # Extract the estimates and populate them in `post_mean_beta`
        start = 0
        self.post_mean_beta = {}

        for c in chroms:
            self.post_mean_beta[c] = res[0][start:start + self.shapes[c]]
            start += self.shapes[c]

        return self



from cython cimport floating

cdef floating[::1] softmax(floating[::1] x) noexcept nogil
cdef floating sigmoid(floating x) noexcept nogil
cdef floating logit(floating x) noexcept nogil
cdef floating dot(floating[::1] v1, floating[::1] v2) noexcept nogil
cdef floating vec_sum(floating[::1] v1) noexcept nogil
cdef void axpy(floating[::1] v1, floating[::1] v2, floating s) noexcept nogil
cdef void scipy_blas_axpy(floating[::1] v1, floating[::1] v2, floating alpha) noexcept nogil
cdef floating scipy_blas_dot(floating[::1] v1, floating[::1] v2) noexcept nogil

cdef floating[::1] clip_list(floating[::1] a, floating min_value, floating max_value) noexcept nogil
cdef floating c_max(floating[::1] x) noexcept nogil
cdef floating clip(floating a, floating min_value, floating max_value) noexcept nogil
